#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from typing import Optional

import casadi as ca

from .base import Controller
from ..base import Base, TimeSeries
from ..dynamic_model.dynamic_model import Model
from ...util.util import convert, is_real, is_symmetric, is_psd, is_pd


class LinearQuadraticRegulator(Controller, Base):
    """
    Class for linear-quadratic regulator (LQR)

    :param model:
    :type model:
    :param id:
    :type id: str, optional
    :param name:
    :type name: str, optional
    :param discrete:
    :type discrete: bool
    :param plot_backend:
    :type plot_backend: str, optional
    """
    def __init__(
            self,
            model: Model,
            id: Optional[str] = None,
            name: Optional[str] = None,
            discrete: bool = True,
            plot_backend: Optional[str] = None
    ) -> None:
        # TODO: Check for empty models
        # NOTE: I don't know how much sense the first if-condition makes, since it's already clear from the typing
        #  information, that we need a Model object.
        if not isinstance(model, Model):
            raise TypeError("The model must be an object of the Model class.")
        if not model.is_setup():
            raise RuntimeError("Model is not set up. Run Model.setup() before passing it to the controller.")
        if discrete and model.continuous:
            raise RuntimeError("The model used for the LQR needs to be discrete. Use Model.discretize() to obtain a "
                               "discrete model.")
        if not discrete and model.discrete:
            raise RuntimeError("The model used for the LQR needs to be continuous.")
        if not model.is_linear():
            raise RuntimeError("The model used for the LQR needs to be linear. Use Model.linearize() to obtain a "
                               "linearized model.")
        if model.is_autonomous():
            raise RuntimeError("The model used for the LQR is autonomous.")

        super().__init__(id=id, name=name)
        if self._id is None:
            self._create_id()

        self._model = model.copy(setup=True)
        self._Q = None
        self._R = None
        self._N = None
        self._K = None
        self._n_x = 0
        self._n_u = 0
        self._n_p = 0
        self._horizon = None

        self._solution = TimeSeries(plot_backend, parent=self)

    def _update_type(self) -> None:
        """

        :return:
        """
        self._type = 'LQR'

    @property
    def Q(self) -> Optional[ca.DM]:
        """

        :return:
        """
        return self._Q

    @Q.setter
    def Q(self, arg):
        if not is_real(arg):
            raise ValueError("LQR matrix Q needs to be real-valued")
        Q = convert(arg, ca.DM)
        if not Q.is_square():
            Q = ca.diag(Q)
        if Q.shape != (self._n_x, self._n_x):
            raise ValueError(f"Dimension mismatch. Supplied dimension is {Q.shape[0]}x{Q.shape[1]}, but required "
                             f"dimension is {self._n_x}x{self._n_x}")
        if not is_symmetric(Q):
            raise ValueError("LQR matrix Q needs to be symmetric")
        if not is_psd(Q):
            raise ValueError("LQR matrix Q needs to be positive semidefinite")
        self._Q = Q
        if self._K is not None:
            self._K = None

    @property
    def R(self) -> Optional[ca.DM]:
        """

        :return:
        """
        return self._R

    @R.setter
    def R(self, arg):
        if not is_real(arg):
            raise ValueError("LQR matrix R needs to be real-valued")
        R = convert(arg, ca.DM)
        if not R.is_square():
            R = ca.diag(R)
        if R.shape != (self._n_u, self._n_u):
            raise ValueError(f"Dimension mismatch. Supplied dimension is {R.shape[0]}x{R.shape[1]}, but required "
                             f"dimension is {self._n_u}x{self._n_u}")
        if not is_symmetric(R):
            raise ValueError("LQR matrix R needs to be symmetric")
        if not is_pd(R):
            raise ValueError("LQR matrix R needs to be positive definite")
        self._R = R
        if self._K is not None:
            self._K = None

    @property
    def N(self) -> Optional[ca.DM]:
        """

        :return:
        """
        return self._N

    @property
    def feedback_gain(self) -> Optional[ca.DM]:
        """

        :return:
        """
        return self._K

    K = feedback_gain

    @property
    def n_x(self) -> int:
        """

        :return:
        """
        return self._n_x

    @property
    def n_u(self) -> int:
        """

        :return:
        """
        return self._n_u

    @property
    def n_p(self) -> int:
        """

        :return:
        """
        return self._n_p

    @property
    def horizon(self) -> Optional[int]:
        """

        :return:
        """
        return self._horizon

    @horizon.setter
    def horizon(self, arg: int) -> None:
        self._horizon = arg

    def setup(self, **kwargs) -> None:
        """

        :param kwargs:
        :return:
        """
        self._n_x = self._model.n_x
        self._n_u = self._model.n_u
        self._n_p = self._model.n_p

        x = ca.SX.sym('x', self._n_x)
        p = ca.SX.sym('p', self._n_p)

        # NOTE: For now only discrete formulations are supported
        if self._horizon is None:
            # Infinite horizon LQR
            raise NotImplementedError("Infinite horizon LQR will be implemented in future releases")
        else:
            # Finite horizon LQR
            # Model matrices
            state_matrix = ca.Function('state_matrix', [self._model.p, self._model.dt, self._model.t], [self._model.A])
            input_matrix = ca.Function('input_matrix', [self._model.p, self._model.dt, self._model.t], [self._model.B])
            # NOTE: For now we ignore time-variant systems (i.e., [] instead of t)
            A = state_matrix(p, self._model.solution.dt, [])  # switch to self._solution if it is used
            B = input_matrix(p, self._model.solution.dt, [])  # switch to self._solution if it is used

            # LQR matrices
            P = ca.SX.sym('P', self._n_x, self._n_x)
            Q = ca.SX.sym('Q', self._n_x, self._n_x)
            R = ca.SX.sym('R', self._n_u, self._n_u)
            N = ca.SX.sym('N', self._n_x, self._n_u)

            # Solve dynamic Riccati equation
            Pk = P
            for k in range(self._horizon):
                APBN = A.T @ Pk @ B + N
                RBPB = R + B.T @ Pk @ B
                Pk = A.T @ Pk @ A - ca.solve(RBPB.T, APBN.T).T @ (B.T @ Pk @ A + N.T) + Q

            # Optimal control input
            K = ca.solve(R + B.T @ Pk @ B, B.T @ Pk @ A + N.T)
            u = -K @ x

            self._function = ca.Function('function',
                                         [x, p, P, Q, R, N],
                                         [u, K],
                                         ['x', 'p', 'P', 'Q', 'R', 'N'],
                                         ['u', 'K'])

            self.check_consistency()

            self._Q = None
            self._R = None
            self._N = ca.DM.zeros(N.shape)
            self._K = None

            # TODO: Set up and use solution?

    def call(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        # TODO: Rename to optimize, so that the syntax is close to the LMPC class
        if self._function is None:
            raise RuntimeError("LQR is not set up. Run LQR.setup(...) before calling the LQR.")
        if self._Q is None:
            raise RuntimeError("Matrix Q is not set properly. To ensure that a unique solution exists, the matrix Q "
                               "needs to be symmetric, real-valued and positive semidefinite.")
        if self._R is None:
            raise RuntimeError("Matrix R is not set properly. To ensure that a unique solutions exists, the matrix R "
                               "needs to be symmetric, real-valued and positive definite.")

        x = kwargs.get('x')
        if x is not None:
            x = convert(x, ca.DM, axis=1)
        else:
            # TODO: Once reference/trajectory tracking and path following are supported we could revert back to allowing
            #  this (and consecutively initializing the states with 0)
            raise ValueError("No state information was supplied to the LQR!")

        p = kwargs.get('p')
        if p is not None:
            p = convert(p, ca.DM, axis=1)
        else:
            p = ca.DM.zeros(1, self._n_p)

        args = {
            'x': x,
            'p': p,
            'P': self._Q,
            'Q': self._Q,
            'R': self._R,
            'N': self._N
        }

        result = self._function(**args)
        if self._K is None:
            self._K = result['K']

        return result['u']


__all__ = [
    'LinearQuadraticRegulator'
]



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from typing import Optional
import warnings

import numpy as np
import torch
from torch import Tensor
from torch.nn import Module, ModuleList, Linear, Dropout, MSELoss, L1Loss, CosineSimilarity, SmoothL1Loss
from torch.nn import functional as F
from torch.optim import Adadelta, Adagrad, Adam, AdamW, SparseAdam, Adamax, ASGD, LBFGS, RMSprop, Rprop, SGD
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.utils.data import Dataset, DataLoader

from ...util.machine_learning import net_to_casadi_graph


class _MLP(Module):
    """Multilayer perceptron"""
    def __init__(self, n_features, n_labels, layers):
        """Constructor method"""
        super().__init__()

        hidden, activation = _process_layers(n_features, n_labels, layers)
        self._hidden = hidden
        self._activation = activation
        self._n_layers = len(layers)
        self._nodes = (n_features, ) + tuple(layer.nodes for layer in layers if layer.type != 'dropout') + (n_labels, )

    @property
    def shape(self):
        """

        :return:
        """
        return self._nodes

    def forward(self, x):
        """

        :param x:
        :return:
        """
        for k in range(self._n_layers):
            x = self._hidden[k](x)
            if self._activation[k] is not None:
                x = self._activation[k](x)
        # NOTE: Right now the output layer is assumed to be linear and cannot be changed
        x = self._hidden[-1](x)

        return x


class _DataSet(Dataset):
    """"""
    def __init__(self, x_data, y_data):
        """Constructor method"""
        self.features = x_data
        self.labels = y_data

    def __len__(self):
        """Length method"""
        return len(self.features)

    def __getitem__(self, item):
        """Item getter method"""
        return self.features[item, :], self.labels[item, :]


class _EarlyStopping:
    """"""
    def __init__(self, folder='.', patience=7, verbose=False, delta=0, debug=False):
        """Constructor method"""
        self._patience = patience
        self._verbose = verbose
        self._counter = 0
        self._best_score = None
        self.early_stop = False
        self._val_loss_min = np.inf
        self._delta = delta
        self._folder = folder
        self._debug = debug

    def __call__(self, val_loss, model):
        """Calling method"""
        score = -val_loss

        if self._best_score is None:
            self._best_score = score
            if not self._debug:
                self._save_checkpoint(val_loss, model)
        elif score < self._best_score + self._delta:
            self._counter += 1
            if self._verbose:
                print(f"Early stopping counter: {self._counter} out of {self._patience}")
            if self._counter >= self._patience:
                self.early_stop = True
        else:
            self._best_score = score
            if not self._debug:
                self._save_checkpoint(val_loss, model)
            self._counter = 0

    def _save_checkpoint(self, val_loss, model):
        """
        Saves model when validation loss decreases

        :param val_loss:
        :param model:
        :return:
        """
        if self._verbose:
            print(f"Test loss decreased ({self._val_loss_min:.6f} --> {val_loss:.6f}). Saving model...")
        torch.save(model.state_dict(), f'{self._folder}/checkpoint.pt')
        self._val_loss_min = val_loss


class MAPELoss(Module):
    """"""
    def __init__(self, dimension: int) -> None:
        """Constructor method"""
        super().__init__()

        self._ones = torch.ones(1, dimension)

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        """

        :param input:
        :param target:
        :return:
        """
        return mape_loss(input, target, ones=self._ones)


class MSLELoss(Module):
    """"""
    def __init__(self, dimension: int) -> None:
        """Constructor method"""
        super().__init__()

        self._ones = torch.ones(1, dimension)

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        """

        :param input:
        :param target:
        :return:
        """
        return msle_loss(input, target, ones=self._ones)


class LogCoshLoss(Module):
    """"""
    def __init__(self, eps: float = 1e-8) -> None:
        """Constructor method"""
        super().__init__()

        self._eps = eps

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        """

        :param input:
        :param target:
        :return:
        """
        return logcosh_loss(input, target, eps=self._eps)


def mape_loss(input: Tensor, target: Tensor, ones: Optional[Tensor] = None, reduction: str = 'mean') -> Tensor:
    """

    :param input:
    :param target:
    :param ones:
    :param reduction:
    :return:
    """
    if ones is None:
        dimension = target.numel()
        ones = torch.ones(1, dimension)
    return torch.mul(F.l1_loss(ones, input / target, reduction=reduction), 100)


def msle_loss(input: Tensor, target: Tensor, ones: Optional[Tensor] = None, reduction: str = 'mean') -> Tensor:
    """

    :param input:
    :param target:
    :param ones:
    :param reduction:
    :return:
    """
    if ones is None:
        dimension = target.numel()
        ones = torch.ones(1, dimension)
    return F.mse_loss(torch.log(input + ones), torch.log(target + ones), reduction=reduction)


def logcosh_loss(input: Tensor, target: Tensor, eps: float = 1e-8, reduction: str = 'mean'):
    """

    :param input:
    :param target:
    :param eps:
    :param reduction:
    :return:
    """
    return torch.mean(torch.log(torch.cosh(target - input + eps)))


def rmse_loss(input: Tensor, target: Tensor, reduction: str = 'mean') -> Tensor:
    """

    :param input:
    :param target:
    :param reduction:
    :return:
    """
    return torch.sqrt(F.mse_loss(input, target, reduction=reduction))


LOSS = {
    'mse': MSELoss,
    'mae': L1Loss,
    'mape': MAPELoss,
    'msle': MSLELoss,
    'cs': CosineSimilarity,
    'huber': SmoothL1Loss,
    'log-cosh': LogCoshLoss
}
OPTIMIZER = {
    'adadelta': Adadelta,
    'adagrad': Adagrad,
    'adam': Adam,
    'adamw': AdamW,
    'sparse_adam': SparseAdam,
    'adamax': Adamax,
    'asgd': ASGD,
    # 'ftrl': NotImplemented,
    'lbfgs': LBFGS,
    # 'nadam': NotImplemented,
    'rmsprop': RMSprop,
    'rprop': Rprop,
    'sgd': SGD
}
METRIC = {
    'mse': F.mse_loss,
    'rmse': rmse_loss,
    'mae': F.l1_loss,
    'mape': mape_loss,
    'msle': msle_loss,
    'cs': F.cosine_similarity,
    'log-cosh': logcosh_loss
}


class _PyTorchWrapper:
    """"""
    def __init__(self, module: Module, **kwargs) -> None:
        """Constructor method"""
        self._module = module

        self._seed = kwargs.get('seed')

        learning_rate = kwargs.get('learning_rate')
        if learning_rate is None:
            learning_rate = .001

        loss = kwargs.get('loss')
        if loss is None:
            loss = 'mse'
        loss = loss.lower()
        if loss in ['mape', 'msle']:
            self._loss = LOSS[loss.lower()](self._module.shape[-1])
        else:
            self._loss = LOSS[loss.lower()]()

        optimizer = kwargs.get('optimizer')
        if optimizer is None:
            optimizer = 'adam'
        self._optimizer = OPTIMIZER[optimizer.lower()](self._module.parameters(), lr=learning_rate)

        metric = kwargs.get('metric')
        if metric is None:
            metric = []
        self._metrics = [METRIC[k.lower()] for k in metric]
        self._metric_names = metric

        device = kwargs.get('device')
        if device is None:
            device = 'cuda:0' if torch.cuda.is_available() else 'cpu'
        self._device = torch.device(device)
        self._dtype = torch.float
        self._module.to(device=self._device, dtype=self._dtype)

        self._tensorboard = kwargs.get('tensorboard')
        self._tensorboard_comment = {
            'batch_size': 0,
            'learning_rate': learning_rate,
            'n_layers': len(self._module.shape) - 2,
            'nodes': self._module.shape[1:-1],
            'seed': self._seed
        }
        self._browser = kwargs.get('browser')

        self._train_loader = None
        self._validate_loader = None
        self._test_loader = None
        self._early_stopping = None
        self._scheduler = None
        self._writer = None

    def _get_data_loader(self, data, batch_size, shuffle):
        """

        :param data:
        :param batch_size:
        :param shuffle:
        :return:
        """
        x_data = torch.tensor(data[0], device=self._device, dtype=self._dtype)
        y_data = torch.tensor(data[1], device=self._device, dtype=self._dtype)
        data_set = _DataSet(x_data, y_data)
        data_loader = DataLoader(data_set, batch_size=batch_size, shuffle=shuffle)
        return data_loader

    def _preprocessing(
            self,
            data,
            validation_data,
            batch_size,
            verbose,
            patience,
            shuffle
    ):
        """

        :param data:
        :param validation_data:
        :param batch_size:
        :param verbose:
        :param patience:
        :param shuffle:
        :return:
        """
        self._train_loader = self._get_data_loader(data, batch_size, shuffle)
        # TODO: Does it make sense to shuffle validation data
        self._test_loader = self._get_data_loader(validation_data, batch_size, False)

        if patience is not None and len(self._test_loader) > 0:
            if verbose > 1:
                self._early_stopping = _EarlyStopping(patience=patience, verbose=True)
            else:
                self._early_stopping = _EarlyStopping(patience=patience)
            self._scheduler = ReduceLROnPlateau(optimizer=self._optimizer, patience=int(patience / 2))
        elif patience is not None and len(self._test_loader) == 0:
            warnings.warn("Early stopping is only available when validation data is supplied for the training of the "
                          "neural network")

        if self._tensorboard is not None:
            self._tensorboard_comment['batch_size'] = batch_size
            self._writer = self._tensorboard.get_summary_writer(**self._tensorboard_comment)
            if self._browser is not None:
                self.show_tensorboard()

    def _train(self, epoch, train_losses, train_metrics, verbose):
        """

        :param epoch:
        :param train_losses:
        :param train_metrics:
        :param verbose:
        :return:
        """
        running_loss = 0.
        running_metric = torch.zeros(len(self._metrics), device=self._device, dtype=self._dtype)

        # TODO: Need a better variable name
        batches = len(self._train_loader)

        self._module.train()

        for idx, (x, y) in enumerate(self._train_loader):
            self._optimizer.zero_grad()
            pred = self._module(x)
            # TODO: Constraint rates to zero
            loss = self._loss(pred, y)
            metric = [metric(pred, y) for metric in self._metrics]
            loss.backward()
            self._optimizer.step()
            running_loss += loss
            running_metric += torch.tensor(metric, device=self._device, dtype=self._dtype)
            if verbose == 1:
                _progress_bar(idx + 1, batches, running_loss / (idx + 1), self._metric_names,
                              running_metric / (idx + 1), 30, "=", "\r")
        if verbose == 1:
            print()

        train_loss = running_loss / batches
        train_metric = running_metric / batches
        train_losses.append(train_loss.detach().cpu().numpy())
        train_metrics.append(train_metric.detach().cpu().numpy())

        if verbose == 2:
            _one_line_per_epoch(batches, train_loss, self._metric_names, train_metric)

        if self._writer is not None:
            self._writer.add_scalar('train_loss', train_loss, epoch)
            for k in range(len(self._metrics)):
                self._writer.add_scalar('train_metric_' + str(k + 1), train_metric[k], epoch)
            for name, param in self._module.named_parameters():
                self._writer.add_histogram(name, param, epoch)
                self._writer.add_histogram(f'{name}.grad', param.grad, epoch)

    def _test(self, data_loader, epoch, test_losses, test_metrics, verbose):
        """

        :param data_loader:
        :param epoch:
        :param test_losses:
        :param test_metrics:
        :param verbose:
        :return:
        """
        running_loss = 0.
        running_metric = torch.zeros(len(self._metrics), device=self._device, dtype=self._dtype)

        # TODO: Need a better variable name
        batches = len(data_loader)

        self._module.eval()

        with torch.no_grad():
            for idx, (x, y) in enumerate(data_loader):
                self._optimizer.zero_grad()
                pred = self._module(x)
                # TODO: Constraint rates to zero
                metric = [metric(pred, y) for metric in self._metrics]
                running_loss += self._loss(pred, y)
                running_metric += torch.tensor(metric, device=self._device, dtype=self._dtype)
                if verbose == 1:
                    _progress_bar(idx + 1, batches, running_loss / (idx + 1), self._metric_names,
                                  running_metric / (idx + 1), 30, "=", "\r")
            if verbose == 1:
                print()

            test_loss = running_loss / batches
            test_metric = running_metric / batches
            test_losses.append(test_loss.detach().cpu().numpy())
            test_metrics.append(test_metric.detach().cpu().numpy())

            if verbose == 2:
                _one_line_per_epoch(batches, test_loss, self._metric_names, test_metric)

            if self._early_stopping is not None:
                self._early_stopping(test_loss, self._module)
                self._scheduler.step(test_loss)

            if self._writer is not None:
                self._writer.add_scalar('test_loss', test_loss, epoch)
                for k in range(len(self._metrics)):
                    self._writer.add_scalar('test_metric_' + str(k + 1), test_metric[k], epoch)

    def train(
            self,
            data,
            validation_data,
            batch_size,
            epochs,
            verbose,
            patience,
            shuffle
    ):
        """

        :param data:
        :param validation_data:
        :param batch_size:
        :param epochs:
        :param verbose:
        :param patience:
        :param shuffle:
        :return:
        """
        self._preprocessing(data, validation_data, batch_size, verbose, patience, shuffle)

        train_losses = []
        train_metrics = []
        test_losses = []
        test_metrics = []

        for epoch in range(epochs):
            if verbose > 0:
                print(f"Epoch {epoch + 1}/{epochs}")

            self._train(epoch, train_losses, train_metrics, verbose)

            if len(self._test_loader) > 1:
                if verbose > 0:
                    print("Evaluate on validation data")
                loss, metrics = self.evaluate(epoch=epoch, verbose=verbose)
                test_losses.extend(loss)
                test_metrics.extend(metrics)

            if self._early_stopping is not None and self._early_stopping.early_stop:
                if verbose > 0:
                    print("Early stopping")
                break

        if self._writer is not None:
            self._writer.flush()
            self._writer.close()

    def evaluate(
            self,
            data=None,
            batch_size=None,
            epoch=1,
            verbose=1
    ):
        """

        :param data:
        :param batch_size:
        :param epoch:
        :param verbose:
        :return:
        """
        if data is not None:
            data_loader = self._get_data_loader(data, batch_size, False)
        else:
            data_loader = self._test_loader

        test_losses = []
        test_metrics = []

        self._test(data_loader, epoch, test_losses, test_metrics, verbose)

        return test_losses, test_metrics

    def build_graph(self, x, layers, input_scaling=None, output_scaling=None):
        """

        :param x:
        :param layers:
        :param input_scaling:
        :param output_scaling:
        :return:
        """
        return net_to_casadi_graph(self, x, layers, input_scaling=input_scaling, output_scaling=output_scaling)

    def get_weights_and_bias(self):
        """

        :return:
        """
        weights = []
        bias = []

        for name, param in self._module.named_parameters():
            if name.endswith('weight'):
                weights.append(param.detach().cpu().numpy())
            elif name.endswith('bias'):
                bias.append(param.detach().cpu().numpy())

        return weights, bias

    def save_net(self, path_to_file):
        """

        :param path_to_file:
        :return:
        """
        if not path_to_file.endswith('.pt'):
            path_to_file += '.pt'
        torch.save(self._module.state_dict(), path_to_file)

    def show_tensorboard(self, browser: Optional[str] = None) -> None:
        """

        :param browser:
        :return:
        """
        if browser is None:
            browser = self._browser
        browser = browser.lower()

        if self._tensorboard is not None:
            self._tensorboard.show(browser)
        else:
            warnings.warn("No TensorBoard was saved. If you want to see a TensorBoard set the corresponding flag "
                          "during setup and train again.")

    def close_tensorboard(self):
        """

        :return:
        """
        if self._tensorboard is not None:
            self._tensorboard.close()


LAYER = {
    'dense': Linear,
    'dropout': Dropout
}
INITIALIZER = {
    'uniform': torch.nn.init.uniform_,
    'normal': torch.nn.init.normal_,
    'constant': torch.nn.init.constant_,
    'ones': torch.nn.init.ones_,
    'zeros': torch.nn.init.zeros_,
    'eye': torch.nn.init.eye_,
    'xavier_uniform': torch.nn.init.xavier_uniform_,
    'xavier_normal': torch.nn.init.xavier_normal_,
    'kaiming_uniform': torch.nn.init.kaiming_uniform_,
    'kaiming_normal': torch.nn.init.kaiming_normal_,
    'orthogonal': torch.nn.init.orthogonal_,
    'sparse': torch.nn.init.sparse_
}
ACTIVATION = {
    'relu': torch.relu,
    'tanh': torch.tanh,
    'sigmoid': torch.sigmoid,
    'softmax': torch.softmax
}


def _process_layers(n_features: int, n_labels: int, layers: list) -> (ModuleList, list):
    """

    :param n_features:
    :param n_labels:
    :param layers:
    :return:
    """
    # NOTE: Right now the output layer is assumed to be linear and cannot be changed
    n_inputs = n_features
    hidden = []
    activation = []
    for layer in layers:
        type_ = layer.type.lower()
        if type_ == 'dropout':
            hidden.append(LAYER[type_](p=layer.rate))
            activation.append(None)
        else:
            layer_ = LAYER[type_](n_inputs, layer.nodes)
            initializer_ = layer.initializer
            if isinstance(initializer_, str):
                initializer = INITIALIZER[initializer_.lower()]
                initializer(layer_.weight)
                initializer(layer_.bias)
            elif isinstance(initializer_, dict):
                key = list(initializer_)[0]
                initializer = INITIALIZER[key]
                val = initializer_[key]
                if isinstance(val, dict):
                    initializer(layer_.weight, **val)
                    initializer(layer_.bias, **val)
                else:
                    initializer(layer_.weight, val)
                    initializer(layer_.bias, val)
            hidden.append(layer_)
            activation.append(ACTIVATION[layer.activation.lower()])
            n_inputs = layer.nodes
    hidden.append(LAYER['dense'](n_inputs, n_labels))
    return ModuleList(hidden), activation


def _progress_bar(iteration, total, loss, metric_names, metrics, length, fill, print_end):
    """

    :param iteration:
    :param total:
    :param loss:
    :param metric_names:
    :param metrics:
    :param length:
    :param fill:
    :param print_end:
    :return:
    """
    metrics_string = ""
    if metric_names:
        for k, metric in enumerate(metrics):
            metrics_string += f" - {metric_names[k]}: {metric:.4f}"
    filled_length = int(length * iteration // total)
    unfilled_length = length - filled_length
    bar = fill * filled_length + ">" * bool(unfilled_length) + "." * (unfilled_length - 1)
    print(f"\r{iteration}/{total} [{bar}] - loss: {loss:.4f}{metrics_string}", end=print_end)


def _one_line_per_epoch(total, loss, metric_names, metrics):
    """

    :param total:
    :param loss:
    :param metric_names:
    :param metrics:
    :return:
    """
    metrics_string = ""
    if metric_names:
        for k, metric in enumerate(metrics):
            metrics_string += f" - {metric_names[k]}: {metric:.4f}"
    print(f"{total}/{total} - loss: {loss:.4f}{metrics_string}")


MODULES = {
    'mlp': _MLP
}


def get_wrapper(kind, *args, **kwargs):
    """

    :param kind:
    :param args:
    :param kwargs:
    :return:
    """
    seed = kwargs.get('seed')
    if seed is not None:
        torch.manual_seed(seed)

    module = MODULES[kind.lower()](*args)

    return _PyTorchWrapper(module, **kwargs)


__all__ = [
    'get_wrapper'
]



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

# NOTE: Adapted from pandas plotting functionality

from .plot import *
from ...util.util import is_list_like


PLOT_CLASSES = {
    'line': LinePlot,
    'dashed': DashedPlot,
    'dotted': DottedPlot,
    'dashdot': DashDotPlot,
    'step': StepPlot,
    'scatter': ScatterPlot
}


def plot(data, kind, **kwargs):
    """

    :param data:
    :param kind:
    :param kwargs:
    :return:
    """
    if is_list_like(kind):
        plot_obj = MultiPlot(data, kind, **kwargs)
    else:
        plot_obj = PLOT_CLASSES[kind](data, **kwargs)
    plot_obj.generate()
    plot_obj.draw()
    return plot_obj.result


__all__ = []



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

import functools
from typing import Any, Callable, Optional, Sequence, TypeVar, Union, cast

import casadi as ca
import numpy as np

from ..modules.object import Object
from ..modules.machine_learning.base import LearningBase
from .probability import Prior
from .util import convert, is_list_like


ML = TypeVar('ML', bound=LearningBase)
Pr = TypeVar('Pr', bound=Prior)
Func = TypeVar('Func', bound=Callable[..., Any])
Numeric = Union[int, float]


class Activation:
    """"""
    def __new__(cls, *args, **kwargs):
        activation = super().__new__(cls)
        activation.__init__(*args)
        return activation()

    def __init__(self, name):
        self._name = name

    def __call__(self, *args, **kwargs):
        return {
            'sigmoid': self.sigmoid,
            'tanh': ca.tanh,
            'relu': self.rectifier,
            'softplus': self.soft_plus,
            'softmax': self.soft_max,
            'linear': self.linear,
            'scale': self.scale
        }[self._name]

    @staticmethod
    def sigmoid(x):
        """

        :param x:
        :return:
        """
        return 1 / (1 + ca.exp(-x))

    @staticmethod
    def rectifier(x):
        """

        :param x:
        :return:
        """
        return ca.fmax(0, x)

    @staticmethod
    def soft_plus(x):
        """

        :param x:
        :return:
        """
        return ca.log(1 + ca.exp(x))

    @staticmethod
    def soft_max(x):
        """

        :param x:
        :return:
        """
        x_exp = ca.exp(x)
        return x_exp / ca.sum1(x_exp)

    @staticmethod
    def linear(x):
        """

        :param x:
        :return:
        """
        return x

    @staticmethod
    def scale(x):
        """

        :param x:
        :return:
        """
        return 1 / (1 - x)


def register_hyperparameters(obj: ML, ids: Sequence[str]):
    """

    :param obj:
    :param ids:
    :return:
    """
    for i in ids:
        HyperparameterManager.register(i, obj)


class HyperparameterManager:
    """
    :note: This class providing a decorator method is probably overkill. We could just as easily let
        register_hyperparameters() populate a global dictionary that is then used by the corresponding properties,
        where the update method of the values is called.
    """
    _mapping = {}

    @classmethod
    def register(cls, key: str, value: ML) -> None:
        """

        :param key:
        :param value:
        :return:
        """
        cls._mapping[key] = value

    @classmethod
    def observe(cls) -> Func:
        """

        :return:
        """
        def decorator(func: Func) -> Func:
            """

            :param func:
            :return:
            """
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                """

                :param args:
                :param kwargs:
                :return:
                """
                name = args[0].name  # First argument is always the Hyperparameter object
                idx = args[0].id
                if idx in cls._mapping:
                    cls._mapping[idx].update(name, *args[1:])
                return func(*args, **kwargs)
            return cast(Func, wrapper)
        return decorator


class Parameter(Object):
    """
    Base class for all parameters

    This class implements the hyperparameters and their behaviour, including the correct setting and getting of their
    values and bounds. A hyperparameter can be flagged as 'fixed' and thereby excluded from any changes during the
    optimization routines in the gaussian process regression.

    :param name: Name of the hyperparameter that serves as identifier in the kernel CasADi function and during updates
    :param value: Single values must always lie inside the bounds. Can contain a single item or multiple, e.g. a vector
        for length scales. Defaults to 1.
    :param prior:
    :param kwargs:
    """
    # TODO: Hyperparameter prior?
    def __init__(
            self,
            name: str,
            value: Union[Numeric, Sequence[Numeric]] = 1.,
            prior: Optional[str, list[str]] = None,
            **kwargs
    ) -> None:
        """Constructor method"""
        super().__init__(name=name)
        self._create_id()

        self._value = convert(value, ca.DM)
        self._shape = self._value.shape

        if prior is not None:
            if not is_list_like(prior):
                prior = [prior]

            n_val = self._value.numel()
            if len(prior) == 1 and n_val != 1:
                prior *= n_val

            if len(prior) > 1:
                raise NotImplementedError
            else:
                params = kwargs.get('prior_parameters')
                prior = self._check_prior(prior[0], params)
        self._prior = prior

        if '.' in name:
            _, name = name.split('.')
        self._SX = ca.SX.sym(name, *self._shape)
        self._MX = ca.MX.sym(name, *self._shape)

        fixed = kwargs.get('fixed')
        if fixed is None:
            fixed = False
        if fixed:
            if prior is not None and prior.name != 'Delta':
                raise RuntimeError("Keyword 'fixed' was set to True and a prior other than 'Delta' was supplied. "
                                   "Please specify one or the other, not both.")
            elif prior is None:
                self._prior = self._check_prior('Delta')

    def __repr__(self) -> str:
        """Representation method"""
        message = "Parameter("
        message += f"name='{self.name}', "
        message += f"value={self._value}, "
        message += f"fixed={self.fixed})"
        return message

    def _values_in_bounds(self) -> bool:  # pragma: no cover
        """

        :return:
        """
        if ca.mmin(ca.fmin(self._value, self._bounds[0])) < self._bounds[0]:
            return False
        if ca.mmax(ca.fmax(self._value, self._bounds[1])) > self._bounds[1]:
            return False
        return True

    @staticmethod
    def _check_prior(name: str, parameters: Optional[dict[str, Union[Numeric, Sequence[Numeric]]]] = None) -> Pr:
        """

        :param name:
        :param parameters:
        :return:
        """
        if parameters is None:
            parameters = {}

        id_ = name.replace("'", "").replace(' ', '_').lower()
        if id_ == 'gaussian':
            prior = Prior.gaussian(**parameters)
        elif id_ == 'laplace':
            prior = Prior.laplace(**parameters)
        elif id_ == 'students_t':
            prior = Prior.students_t(**parameters)
        elif id_ == 'delta':
            prior = Prior.delta()
        else:
            raise ValueError(f"Prior '{name}' not recognized")

        return prior

    @HyperparameterManager.observe()
    def _set_properties(self, value, bounds):
        """

        :param value:
        :param bounds:
        :return:
        """
        # TODO: Check whether it could happen that both value and bounds are not None
        if value is not None:
            # NOTE: This is somewhat of a hack, but since we use floats the optimization may return values just so
            #  slightly out of bounds
            # TODO: Use epsilon and absolute of difference?
            value = convert(value, ca.DM)

            bounds_supplied = True
            if bounds is None:
                bounds = (-ca.inf, ca.inf)
                bounds_supplied = False

            if bool(ca.logic_all(value >= bounds[0])) and bool(ca.logic_all(value <= bounds[1])):
                self._value = value
                new_shape = value.shape
                if new_shape != self._shape:
                    self._shape = new_shape
                    name = self.name
                    if '.' in name:
                        _, name = name.split('.')
                    self._SX = ca.SX.sym(name, *self._shape)
                    self._MX = ca.MX.sym(name, *self._shape)
            else:
                raise ValueError(f"Given {self.name} value is not in between bounds!")

            if not bounds_supplied:
                bounds = None

        if bounds is not None:  # pragma: no cover
            if bounds == 'fixed':
                self._fixed = True
                self._bounds = (0, ca.inf)
            elif isinstance(bounds, (int, float)):
                self._bounds = (bounds, ca.inf)
                self._fixed = False
            elif len(bounds) == 2:
                assert np.less(bounds[0], bounds[1]), "Lower bound not smaller than upper bound!"
                self._bounds = tuple(bounds)
                self._fixed = False
            elif len(bounds) == 1:
                self._bounds = (*bounds, ca.inf)
                self._fixed = False
            else:
                raise ValueError(f"Unsupported type {type(bounds).__name__} for bounds attribute")

    @property
    def value(self) -> ca.DM:
        """
        Single values must always lie inside the bounds. Can contain a single item or multiple, e.g. a vector for
        length scales

        :return:
        """
        return self._value

    @value.setter
    def value(self, value: list[Numeric]) -> None:
        self._set_properties(value, None)

    @property
    def prior(self) -> Pr:
        """

        :return:
        """
        return self._prior

    @prior.setter
    def prior(self, name: str) -> None:
        self._prior = self._check_prior(name)

    @property
    def bounds(self) -> (Numeric, Numeric):  # pragma: no cover
        """
        First item represents the lower bound, the second item the upper bound. No lower/upper bound is defined by
        an inf object.

        :return:
        """
        return self._bounds

    @bounds.setter
    def bounds(self, bounds: Union[str, Numeric, tuple[Numeric], tuple[Numeric, Numeric]]) -> None:  # pragma: no cover
        self._set_properties(self._log, bounds)

    @property
    def lower_bound(self) -> Numeric:  # pragma: no cover
        """
        Lower bound of the hyperparameter value

        :return:
        """
        return self.bounds[0]

    @lower_bound.setter
    def lower_bound(self, lower_bound: Numeric) -> None:  # pragma: no cover
        self.bounds = (lower_bound, self.upper_bound)

    @property
    def upper_bound(self) -> Numeric:  # pragma: no cover
        """
        Upper bound of the hyperparameter value

        :return:
        """
        return self.bounds[1]

    @upper_bound.setter
    def upper_bound(self, upper_bound: Numeric) -> None:  # pragma: no cover
        self.bounds = (self.lower_bound, upper_bound)

    @property
    def SX(self) -> ca.SX:
        """

        :return:
        """
        return self._SX

    @property
    def MX(self) -> ca.MX:  # pragma: no cover
        """

        :return:
        """
        return self._MX

    @property
    def fixed(self) -> bool:
        """

        :return:
        """
        if self._prior is not None:
            return self._prior.name == 'Delta'
        return False

    @fixed.setter
    def fixed(self, value: bool) -> None:
        if value:
            if self._prior is None:
                self._prior = Prior.delta()
            elif self._prior.name != 'Delta':
                raise RuntimeError("Hyperparameter cannot be set to fixed, since a hyperprior already exists")
        else:
            if self._prior is not None:
                if self._prior.name != 'Delta':
                    raise RuntimeError("Hyperparameter cannot be set to variable, since a hyperprior already exists")
                else:
                    self._prior = None

    def is_scalar(self) -> bool:
        """

        :return:
        """
        return self.SX.is_scalar()


class PositiveParameter(Parameter):
    """"""
    def __init__(
            self,
            name: str,
            value: Union[Numeric, Sequence[Numeric]] = 1.,
            prior: Optional[str, list[str]] = None,
            prior_mean: Optional[Numeric, list[Numeric]] = None,
            prior_variance: Optional[Numeric, list[Numeric]] = None,
            bounds: tuple[float, float] = (0., ca.inf),
            fixed: bool = False
    ) -> None:
        """Constructor method"""
        if bounds[0] < 0. or bounds[1] < 0.:  # pragma: no cover
            raise ValueError("Bounds need to be positive")
        super().__init__(name, value=value, prior=prior, prior_mean=prior_mean, prior_variance=prior_variance,
                         bounds=bounds, fixed=fixed)

        if ca.mmin(self._value) <= 0.:
            raise ValueError("Hyperparameters can only take positive values")

        self._log = ca.log(self._value)

    def _set_properties(self, value, bounds):
        """

        :param value:
        :param bounds:
        :return:
        """
        super()._set_properties(value, bounds)

        if value is not None:
            self._log = ca.log(value)

        if bounds is not None:  # pragma: no cover
            self._log_bounds = (ca.log(self._bounds[0]), ca.log(self._bounds[1]))

    @property
    def log(self) -> ca.DM:
        """

        :return:
        """
        return self._log

    @property
    def log_bounds(self) -> (Numeric, Numeric):  # pragma: no cover
        """

        :return:
        """
        return self._log_bounds


Param = TypeVar('Param', bound=Parameter)


class Hyperparameter:
    """"""
    def __new__(cls, *args, **kwargs) -> Param:
        """Creator method"""
        positive = kwargs.pop('positive', None)
        if positive is None:
            positive = True

        if positive:
            return PositiveParameter(*args, **kwargs)
        else:
            return Parameter(*args, **kwargs)


def net_to_casadi_graph(net, x, layers, **kwargs):
    """

    :param net:
    :param x:
    :param layers:
    :param kwargs:
    :return:
    """
    input_scaling = kwargs.get('input_scaling')
    output_scaling = kwargs.get('output_scaling')
    h = x
    graph = []
    # TODO: What if bias was deactivated?
    if isinstance(net, dict):
        weights = net['weights']
        bias = net['bias']
    else:
        weights, bias = net.get_weights_and_bias()

    if input_scaling is not None:
        h -= input_scaling.mean_
        h /= input_scaling.scale_

    # NOTE: We could also use a counter (initialized with 0) here, that increases by 1 every time we come across a
    #  Dropout layer and then subtract the counter from 'k' in every index.
    layers = [layer for layer in layers if layer.type.lower() != 'dropout']
    for k, layer in enumerate(layers):
        if layer.type.lower() == 'dense':
            if not graph:
                y = weights[k] @ h + bias[k]
            else:
                y = weights[k] @ graph[-1] + bias[k]
            graph.append(Activation(layer.activation)(y))
    # NOTE: Right now the output layer is assumed to be linear and cannot be changed
    if not graph:
        y = weights[-1] @ h + bias[-1]
    else:
        y = weights[-1] @ graph[-1] + bias[-1]

    if output_scaling is not None:
        y *= output_scaling.scale_
        y += output_scaling.mean_

    graph = [y] + graph

    hidden = []
    ct = 1
    for layer in layers:
        if layer.type.lower() != 'dropout':
            hidden.append("layer_" + str(ct))
            ct += 1

    return ca.Function('neural_network',
                       [x],
                       graph,
                       ["features"],
                       ["labels"] + hidden)



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

import os
import subprocess
from typing import Optional


UNIX_COMPILERS = ['gcc', 'g++']


def compile_so(path_to_file: str, compiler: str, output: Optional[str] = None) -> str:
    """

    :param path_to_file:
    :param compiler:
    :param output:
    :return:
    """
    if os.path.exists(path_to_file):
        command = [compiler, '-fPIC', '-shared', path_to_file, '-o']
        if output is not None:
            so_path = output
        else:
            so_path = path_to_file.rsplit('.', 1)
            so_path[-1] = 'so'
            so_path = '.'.join(so_path)
        command.append(so_path)
        if subprocess.call(command) == 0:
            return so_path
        else:
            raise RuntimeError("Could not compile library")
    else:
        raise FileNotFoundError(f"File {path_to_file} does not exist")


def find_compiler(compiler: str) -> bool:
    """

    :param compiler:
    :return:
    """
    if subprocess.call([compiler, '--version'], stdout=subprocess.DEVNULL) == 0:
        return True
    else:
        return False



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from abc import ABCMeta, abstractmethod
from typing import Dict, Optional, TypeVar, Union
import warnings

import casadi as ca
import numpy as np

from .likelihood import Likelihood
from .mean import Mean
from .kernel import Kernel


Numeric = Union[int, float]
Symbolic = TypeVar('Symbolic', ca.SX, ca.MX)
Array = TypeVar('Array', ca.SX, ca.MX, np.ndarray)
Lik = TypeVar('Lik', bound=Likelihood)
Mu = TypeVar('Mu', bound=Mean)
Cov = TypeVar('Cov', bound=Kernel)


class Inference(metaclass=ABCMeta):
    """"""
    def __init__(self):
        """Constructor method"""
        self._posterior = {
            'mean': ca.SX(),
            'var': ca.SX(),
            'log_marginal_likelihood': ca.SX()
        }

    def __call__(self, *args, **kwargs) -> Optional[Dict[str, Symbolic]]:
        """Calling method"""
        if args and kwargs:
            raise TypeError("Call to Inference can only accept positional or keyword arguments, not both")

        if args:
            X = args[0]
            y = args[1]
            x_test = args[2]  # test input
            noise_variance = args[3]
            likelihood = args[4]
            mean = args[5]
            kernel = args[6]
        elif kwargs:
            X = kwargs.get('X')
            y = kwargs.get('y')
            x_test = kwargs.get('x_test')
            noise_variance = kwargs.get('noise_variance')
            likelihood = kwargs.get('likelihood')
            mean = kwargs.get('mean')
            kernel = kwargs.get('kernel')

            if X is None:
                raise ValueError("No vector X was supplied")
            if y is None:
                raise ValueError("No vector y was supplied")
            if x_test is None:
                raise ValueError("No test input was supplied")
            if noise_variance is None:
                raise ValueError("No noise variance was supplied")
            if likelihood is None:
                raise ValueError("No likelihood was supplied")
            if mean is None:
                raise ValueError("No mean function was supplied")
            if kernel is None:
                raise ValueError("No covariance function (kernel) was supplied")
        else:
            warnings.warn("No arguments supplied")
            return

        self.get_posterior(X, y, x_test, noise_variance, likelihood, mean, kernel)
        return self._posterior

    @property
    def posterior(self) -> Dict[str, Symbolic]:
        """

        :return:
        """
        return self._posterior

    @abstractmethod
    def get_posterior(
            self,
            X: Array,
            y: Array,
            x_test: Array,
            noise_variance: Union[Symbolic, Numeric],
            likelihood: Lik,
            mean: Mu,
            kernel: Cov
    ) -> None:
        """

        :param X:
        :param y:
        :param x_test:
        :param noise_variance:
        :param likelihood:
        :param mean:
        :param kernel:
        :return:
        """
        pass

    @staticmethod
    def exact():
        """

        :return:
        """
        return ExactInference()

    @staticmethod
    def laplace():
        """

        :return:
        """
        return Laplace()

    @staticmethod
    def expectation_propagation():
        """

        :return:
        """
        return ExpectationPropagation()

    @staticmethod
    def variational_bayes():
        """

        :return:
        """
        return VariationalBayes()

    @staticmethod
    def kullback_leibler():
        """

        :return:
        """
        return KullbackLeibler()


class ExactInference(Inference):
    """"""
    def get_posterior(
            self,
            X: Array,
            y: Array,
            x_test: Array,
            noise_variance: Union[Symbolic, Numeric],
            likelihood: Lik,
            mean: Mu,
            kernel: Cov
    ) -> None:
        """

        :param X:
        :param y:
        :param x_test:
        :param noise_variance:
        :param likelihood:
        :param mean:
        :param kernel:
        :return:
        """
        if likelihood.name != "Gaussian":
            raise ValueError("Exact inference is only applicable with Gaussian likelihood. Choose a different inference"
                             " method in order to use other likelihoods.")

        n, D = X.shape

        noise_variance = ca.exp(2 * noise_variance)

        K = kernel(X, X)
        prior_mu = mean(X)

        # NOTE: CasADi's Cholesky decomposition returns an upper triangular matrix, but we need a lower triangular
        #  matrix, so we transpose 'L' in the calculation of 'alpha'
        L = ca.chol(noise_variance * ca.SX.eye(D) + K)
        y_minus_prior = y - prior_mu
        alpha = ca.solve(L, ca.solve(L.T, y_minus_prior.T))  # \alpha = L^T\(L\y) (see Rasmussen p.19)

        log_marginal_likelihood = -.5 * y_minus_prior @ alpha - ca.sum1(ca.log(ca.diag(L))) - D / 2 * ca.log(2 * ca.pi)

        K = kernel(X, x_test)
        mu = mean(x_test) + K.T @ alpha

        v = ca.solve(L.T, K)  # v = L\k_* (see Rasmussen p.19)
        K = kernel(x_test, x_test)
        var = K - v.T @ v

        self._posterior['mean'] = mu
        self._posterior['var'] = var
        self._posterior['log_marginal_likelihood'] = log_marginal_likelihood


class Laplace(Inference):
    """"""
    def __init__(self):
        """Constructor method"""
        super().__init__()

        raise NotImplementedError("Laplace inference not yet implemented")


class ExpectationPropagation(Inference):
    """"""
    def __init__(self):
        """Constructor method"""
        super().__init__()

        raise NotImplementedError("Expectation propagation inference not yet implemented")


class VariationalBayes(Inference):
    """"""
    def __init__(self):
        """Constructor method"""
        super().__init__()

        raise NotImplementedError("Variational Bayes inference not yet implemented")


class KullbackLeibler(Inference):
    """"""
    def __init__(self):
        """Constructor method"""
        super().__init__()

        raise NotImplementedError("Kullback-Leibler inference not yet implemented")



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

import hilo_mpc.modules.dynamic_model.dynamic_model as dyn_mod
import hilo_mpc.modules.controller.mpc as mpc
import hilo_mpc.modules.controller.ocp as ocp
import hilo_mpc.modules.controller.lqr as lqr
import hilo_mpc.modules.controller.pid as pid
import hilo_mpc.modules.estimator.mhe as mhe
import hilo_mpc.modules.estimator.kf as kf
import hilo_mpc.modules.estimator.pf as pf
import hilo_mpc.modules.machine_learning.nn.layer as layer
import hilo_mpc.modules.machine_learning.nn.nn as nn
import hilo_mpc.modules.machine_learning.gp.mean as mean
import hilo_mpc.modules.machine_learning.gp.kernel as kernel
import hilo_mpc.modules.machine_learning.gp.gp as gp
import hilo_mpc.modules.control_loop as cl
import hilo_mpc.modules.optimizer as opti
import hilo_mpc.util.plotting as plotting
import hilo_mpc.util.session as session


Model = dyn_mod.Model
NMPC = mpc.NMPC
LMPC = mpc.LMPC
SMPC = mpc.SMPC
OptimalControlProblem = ocp.OptimalControlProblem
OCP = OptimalControlProblem
LinearQuadraticRegulator = lqr.LinearQuadraticRegulator
LQR = LinearQuadraticRegulator
PID = pid.PID
MovingHorizonEstimator = mhe.MovingHorizonEstimator
MHE = MovingHorizonEstimator
KalmanFilter = kf.KalmanFilter
KF = KalmanFilter
ExtendedKalmanFilter = kf.ExtendedKalmanFilter
EKF = ExtendedKalmanFilter
UnscentedKalmanFilter = kf.UnscentedKalmanFilter
UKF = UnscentedKalmanFilter
ParticleFilter = pf.ParticleFilter
PF = ParticleFilter
Layer = layer.Layer
Dense = layer.Dense
Dropout = layer.Dropout
ArtificialNeuralNetwork = nn.ArtificialNeuralNetwork
ANN = ArtificialNeuralNetwork
Mean = mean.Mean
ConstantMean = mean.ConstantMean
ZeroMean = mean.ZeroMean
OneMean = mean.OneMean
PolynomialMean = mean.PolynomialMean
LinearMean = mean.LinearMean
Kernel = kernel.Kernel
ConstantKernel = kernel.ConstantKernel
SquaredExponentialKernel = kernel.SquaredExponentialKernel
MaternKernel = kernel.MaternKernel
ExponentialKernel = kernel.ExponentialKernel
Matern32Kernel = kernel.Matern32Kernel
Matern52Kernel = kernel.Matern52Kernel
RationalQuadraticKernel = kernel.RationalQuadraticKernel
PiecewisePolynomialKernel = kernel.PiecewisePolynomialKernel
DotProductKernel = kernel.DotProductKernel
PolynomialKernel = kernel.PolynomialKernel
LinearKernel = kernel.LinearKernel
NeuralNetworkKernel = kernel.NeuralNetworkKernel
PeriodicKernel = kernel.PeriodicKernel
GaussianProcess = gp.GaussianProcess
GP = GaussianProcess
GPArray = gp.GPArray
SimpleControlLoop = cl.SimpleControlLoop
LinearProgram = opti.LinearProgram
LP = LinearProgram
QuadraticProgram = opti.QuadraticProgram
QP = QuadraticProgram
NonlinearProgram = opti.NonlinearProgram
NLP = NonlinearProgram
Session = session.Session

get_plot_backend = plotting.get_plot_backend
set_plot_backend = plotting.set_plot_backend


try:
    # Prefer resolving version from installed package metadata (works with Poetry builds)
    try:
        from importlib.metadata import version, PackageNotFoundError  # Python 3.8+
    except ImportError:  # pragma: no cover - for very old Python
        from importlib_metadata import version, PackageNotFoundError  # type: ignore

    try:
        __version__ = version("hilo-mpc")
    except PackageNotFoundError:
        # Fallback to setuptools_scm write-to file if present (editable installs)
        from . import _version  # type: ignore
        __version__ = getattr(_version, "version", '0.0.0+unknown')
except Exception:  # pragma: no cover
    __version__ = '0.0.0+unknown'


__all__ = [
    'Model',
    'NMPC',
    'LMPC',
    'OptimalControlProblem',
    'OCP',
    'LinearQuadraticRegulator',
    'LQR',
    'PID',
    'MovingHorizonEstimator',
    'MHE',
    'KalmanFilter',
    'KF',
    'ExtendedKalmanFilter',
    'EKF',
    'UnscentedKalmanFilter',
    'UKF',
    'ParticleFilter',
    'PF',
    'Layer',
    'Dense',
    'Dropout',
    'ArtificialNeuralNetwork',
    'ANN',
    'Mean',
    'ConstantMean',
    'ZeroMean',
    'OneMean',
    'PolynomialMean',
    'LinearMean',
    'Kernel',
    'ConstantKernel',
    'SquaredExponentialKernel',
    'MaternKernel',
    'ExponentialKernel',
    'Matern32Kernel',
    'Matern52Kernel',
    'RationalQuadraticKernel',
    'PiecewisePolynomialKernel',
    'DotProductKernel',
    'PolynomialKernel',
    'LinearKernel',
    'NeuralNetworkKernel',
    'PeriodicKernel',
    'GaussianProcess',
    'GP',
    'GPArray',
    'SimpleControlLoop',
    'LinearProgram',
    'LP',
    'QuadraticProgram',
    'QP',
    'NonlinearProgram',
    'NLP',
    'Session',
    'get_plot_backend',
    'set_plot_backend'
]



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

# TODO: Typing hints
# NOTE: Adapted from pandas plotting functionality

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from math import ceil
import pathlib
import warnings

import bokeh.palettes as cm
import numpy as np

from ...util.util import is_list_like, random_state


_LINE_KWARGS = ['line_alpha', 'line_cap', 'line_color', 'line_dash', 'line_dash_offset', 'line_join', 'line_width']
_STEP_KWARGS = ['step_mode']
_SCATTER_KWARGS = ['marker', 'marker_color', 'marker_size']


class BasePlot(metaclass=ABCMeta):
    """"""
    # TODO: Support for format string
    _layout_type = 'grid'
    _pop_attributes = [
        'axis_label_text_font_size',
        'major_label_text_font_size',
        'background_fill_color',
        'logy',
        'logx',
        'loglog'
    ]
    _attr_defaults = {
        'axis_label_text_font_size': None,
        'major_label_text_font_size': None,
        'background_fill_color': None,
        'logy': False,
        'logx': False,
        'loglog': False
    }

    def __init__(
            self,
            data,
            subplots=False,
            figsize=None,
            legend=False,
            legend_title=None,
            ax=None,
            sources=None,
            title=None,
            xlabel=None,
            ylabel=None,
            colormap=None,
            layout=None,
            fill_between=None,
            interactive=False,
            server=False,
            output_file=None,
            output_notebook=False,
            browser=None,
            return_html=False,
            **kwargs
    ):
        """Constructor method"""
        self.data = data
        if sources is None:
            sources = []
        self.sources = sources

        self.subplots = subplots

        self.figsize = figsize
        self.layout = layout
        self.fill_between = fill_between
        self.interactive = interactive
        self.server = server

        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel

        self.legend = legend
        self.legend_title = legend_title

        for attr in self._pop_attributes:
            value = kwargs.pop(attr, self._attr_defaults.get(attr))
            setattr(self, attr, value)

        self.ax = ax

        if 'cmap' in kwargs and colormap:
            raise TypeError("Only specify one of 'cmap' and 'colormap'")
        elif 'cmap' in kwargs:
            self.colormap = kwargs.pop('cmap')
        else:
            self.colormap = colormap

        self.output_file = output_file
        self.output_notebook = output_notebook
        self.browser = browser
        self.return_html = return_html

        self.kwargs = kwargs

    def _decorate_subplots(self):
        """

        :return:
        """
        if is_list_like(self.xlabel):
            if len(self.xlabel) != self.n_plots:
                msg = f"The length of 'xlabel' must equal the number of columns if using 'xlabel' of type " \
                      f"'list'.\nlength of xlabel = {len(self.xlabel)}\nnumber of columns = {self.n_plots}"
                raise ValueError(msg)

        if is_list_like(self.ylabel):
            if len(self.ylabel) != self.n_plots:
                msg = f"The length of 'ylabel' must equal the number of columns if using 'ylabel' of type " \
                      f"'list'.\nlength of ylabel = {len(self.ylabel)}\nnumber of columns = {self.n_plots}"
                raise ValueError(msg)

        for k in range(self.n_plots):
            ax = self._get_glyph(k)
            if ax.visible:
                if is_list_like(self.xlabel):
                    if self.xlabel[k] is not None:
                        ax.xaxis.axis_label = self.xlabel[k]
                else:
                    if self.xlabel is not None:
                        ax.xaxis.axis_label = self.xlabel

                if is_list_like(self.ylabel):
                    if self.ylabel[k] is not None:
                        ax.yaxis.axis_label = self.ylabel[k]
                else:
                    if self.ylabel is not None:
                        ax.yaxis.axis_label = self.ylabel

                if self.axis_label_text_font_size is not None:
                    ax.yaxis.axis_label_text_font_size = self.axis_label_text_font_size
                    ax.xaxis.axis_label_text_font_size = self.axis_label_text_font_size

                if self.major_label_text_font_size is not None:
                    ax.yaxis.major_label_text_font_size = self.major_label_text_font_size
                    ax.xaxis.major_label_text_font_size = self.major_label_text_font_size

            if self.interactive:
                slider = self._get_glyph(k, get_interactive=True)
                _update_slider(slider, self.sources[k])

        if self.title is not None:
            if self.subplots:
                if is_list_like(self.title):
                    if len(self.title) != self.n_plots:
                        msg = f"The length of 'title' must equal the number of columns if using 'title' of type " \
                              f"'list' and 'subplots=True'.\nlength of title = {len(self.title)}\nnumber of columns =" \
                              f" {self.n_plots}"
                        raise ValueError(msg)

                    for k in range(self.n_plots):
                        ax = self._get_glyph(k)
                        ax.title.text = self.title[k]
                else:
                    warnings.warn("An overall title for all subplots is currently not supported")
            else:
                if is_list_like(self.title):
                    msg = "Using 'title' of type 'list' is not supported unless 'subplots=True' is passed."
                    raise ValueError(msg)
                if self.interactive:
                    # NOTE: At the moment the first entry will be the Figure and the second one the Slider
                    self.axes.children[0].title.text = self.title
                else:
                    self.axes.title.text = self.title

    def _get_colors(self, num_colors=None, color_kwargs='color', subplot=None):
        """

        :param num_colors:
        :param color_kwargs:
        :param subplot:
        :return:
        """
        if num_colors is None:
            num_colors = self.n_plots

        if subplot is None:
            color = self.kwargs.get(color_kwargs)
        else:
            color = self.kwargs.get(color_kwargs)[subplot]

        return _get_standard_colors(num_colors=num_colors, colormap=self.colormap, color=color)

    def _get_glyph(self, k, get_interactive=False):
        """

        :param k:
        :param get_interactive:
        :return:
        """
        if self.subplots:
            ax = self.axes.children
            # Convert tuple to list for consistent indexing (Bokeh 3.x compatibility)
            if isinstance(ax, tuple):
                ax = list(ax)
            if len(ax) == 2 and self.n_plots != 2:
                children = ax[1].children
                if isinstance(children, tuple):
                    children = list(children)
                child_k = children[k]
                if isinstance(child_k, tuple):
                    child_k = list(child_k)
                ax = child_k[0]
            elif len(ax) >= self.n_plots:
                ax = ax[k]
                # In Bokeh 3.x, ax[k] might itself be a tuple
                if isinstance(ax, tuple):
                    ax = list(ax)
                    if len(ax) > 0:
                        ax = ax[0]
            else:
                raise ValueError("Number of desired plots is not equal to the number of created plots")
        else:
            ax = self.axes

        if self.interactive:
            if not get_interactive:
                children = ax.children
                if isinstance(children, tuple):
                    children = list(children)
                ax = children[0]
            else:
                children = ax.children
                if isinstance(children, tuple):
                    children = list(children)
                ax = children[1]
                return ax

        ax.yaxis.visible = True
        return ax

    def _iter_data(self):
        """

        :return:
        """
        for sub, plots in self.data.items():
            if len(self.sources) < self.n_plots:
                source_data = {}
                for val in plots['x'].values():
                    if not source_data:
                        source_data['x'] = val
                    else:
                        if val != source_data['x']:
                            raise ValueError
                source_data.update({'y' + str(k): val for k, val in enumerate(plots['y'].values())})
                source = self.source(source_data)
                self.sources.append(source)
            elif len(self.sources) == self.n_plots:
                source = self.sources[sub]
            else:
                raise RuntimeError("length of ColumnDataSource list > number of plots")
            if 'line_color' in self.kwargs:
                colors = self._get_colors(num_colors=len(plots['y']), color_kwargs='line_color')
            elif 'marker_color' in self.kwargs:
                colors = self._get_colors(num_colors=len(plots['y']), color_kwargs='marker_color')
            else:
                colors = self._get_colors(num_colors=len(plots['y']))
            keys = list(plots['y'])
            for k, key in enumerate(keys):
                if isinstance(colors, dict):
                    color = colors[key]
                else:
                    color = colors[keys.index(key)]
                yield sub, k, key, source, color

    def _make_fill_between(self):
        """

        :return:
        """
        if not self.subplots:
            if is_list_like(self.fill_between):
                fill_between = self.fill_between[0]
            else:
                fill_between = self.fill_between
            if fill_between is not None:
                ax = self._get_glyph(0)
                self._process_fill_between(fill_between)
                ax.patch(**fill_between)
        else:
            if is_list_like(self.fill_between):
                if len(self.fill_between) != self.n_plots:
                    msg = f"The length of 'fill_between' must equal the number of columns if using 'fill_between' of " \
                          f"type 'list'.\nlength of fill_between = {len(self.fill_between)}\nnumber of columns = " \
                          f"{self.n_plots}"
                    raise ValueError(msg)
            for k in range(self.n_plots):
                if is_list_like(self.fill_between):
                    fill_between = self.fill_between[k]
                elif self.fill_between is not None:
                    fill_between = self.fill_between.copy()
                else:
                    fill_between = None
                if fill_between is not None:
                    ax = self._get_glyph(k)
                    self._process_fill_between(fill_between)
                    ax.patch(**fill_between)

    def _make_legend(self):
        """

        :return:
        """
        if not self.subplots:
            if is_list_like(self.legend):
                legend = self.legend[0]
            else:
                legend = self.legend
            if not legend:
                self.axes.legend.visible = False
            else:
                if self.legend_title is not None:
                    if is_list_like(self.legend_title):
                        msg = "Using 'legend_title' of type 'list' is not supported unless 'subplots=True' is passed."
                        raise ValueError(msg)
                    self.axes.legend.title = self.legend_title
        else:
            if is_list_like(self.legend):
                if len(self.legend) != self.n_plots:
                    msg = f"The length of 'legend' must equal the number of columns if using 'legend' of type " \
                          f"'list'.\nlength of legend = {len(self.legend)}\nnumber of columns = {self.n_plots}"
                    raise ValueError(msg)
            if is_list_like(self.legend_title):
                if len(self.legend_title) != self.n_plots:
                    msg = f"The length of 'legend_title' must equal the number of columns if using 'legend_title'" \
                          f" of type 'list' and 'legend=True'.\nlength of legend_title = {len(self.legend_title)}" \
                          f"\nnumber of columns = {self.n_plots}"
                    raise ValueError(msg)
            for k in range(self.n_plots):
                ax = self._get_glyph(k)
                if not self.legend or (is_list_like(self.legend) and not self.legend[k]):
                    ax.legend.visible = False
                else:
                    if is_list_like(self.legend_title):
                        ax.legend.title = self.legend_title[k]
                    else:
                        ax.legend.title = self.legend_title

    @abstractmethod
    def _make_plot(self):
        """

        :return:
        """
        pass

    def _process_fill_between(self, kwargs):
        """

        :param kwargs:
        :return:
        """
        x = kwargs.pop('x', None)
        if x is None:
            raise KeyError("No base for 'fill_between' option supplied")
        if x.ndim > 1:
            if 1 in x.shape:
                x = x.flatten()
            else:
                raise ValueError("Base argument for the 'fill_between' option needs to be a flat array")
        lb = kwargs.pop('lb', None)
        if lb is None:
            raise KeyError("No lower bound for 'fill_between' option supplied")
        if lb.ndim > 1:
            if 1 in lb.shape:
                lb = lb.flatten()
            else:
                raise ValueError("Lower bound argument for the 'fill_between' option needs to be a flat array")
        ub = kwargs.pop('ub', None)
        if ub is None:
            raise KeyError("No upper bound for 'fill_between' option supplied")
        if ub.ndim > 1:
            if 1 in ub.shape:
                ub = ub.flatten()
            else:
                raise ValueError("Upper bound argument for the 'fill_between' option needs to be a flat array")

        if x.size == lb.size == ub.size:
            x_lb = x
            x_ub = x_lb[::-1]
            source = self.source({'x': np.hstack((x_lb, x_ub)), 'y': np.hstack((lb, ub[::-1]))})
            kwargs['x'] = 'x'
            kwargs['y'] = 'y'
            kwargs['source'] = source
            if 'label' in kwargs:
                kwargs['legend_label'] = kwargs.pop('label')
        else:
            raise ValueError(f"Dimension mismatch between base, lower and upper bound for the 'fill_between' option."
                             f"\nbase: {x.size}\nlower: {lb.size}\nupper: {ub.size}")

    def _setup_subplots(self):
        """

        :return:
        """
        fig_kwargs = {}
        if self.figsize is not None:
            if self.background_fill_color is not None:
                fig_kwargs['background_fill_color'] = self.background_fill_color

            if isinstance(self.figsize, (list, tuple, set)) and len(self.figsize) <= 2:
                if len(self.figsize) == 1:
                    width = height = self.figsize[0]
                else:
                    width = self.figsize[0]
                    height = self.figsize[1]
                fig_kwargs['width'] = width
                fig_kwargs['height'] = height

        # TODO: Initialize slider here already and set properties later
        if self.subplots:
            # TODO: Add support for self.ax
            if self.interactive:
                fig_kwargs['interactive'] = True
                fig_kwargs['x_data'] = [data['x'] for data in self.data.values()]
            fig = _subplots(n_axes=self.n_plots, subplot_kwargs=fig_kwargs, layout=self.layout,
                            layout_type=self._layout_type)
        else:
            if self.ax is None:
                fig = self.plt.figure(**fig_kwargs)
            else:
                fig = self.ax
            if self.interactive:
                fig = self.column(fig, _slider(self.data[0]['x']))

        self.axes = fig

    @property
    def column(self):
        """

        :return:
        """
        from bokeh.layouts import column

        return column

    @property
    def n_plots(self):
        """

        :return:
        """
        return len(self.data)

    @property
    def plt(self):
        """

        :return:
        """
        import bokeh.plotting as plt

        return plt

    @property
    def result(self):
        """

        :return:
        """
        if self.server:
            from bokeh.models import Button
            from bokeh.server.server import Server
            # from bokeh.themes import Theme

            # path = pathlib.Path(__file__)
            # file = str(path.parent.absolute()) + '/theme.yaml'

            return self.column(Button(), self.axes), self.sources, Server, None  # Theme(filename=file)
        elif self.return_html:
            from bokeh.resources import CDN
            from bokeh.embed import file_html

            return file_html(self.axes, CDN)
        else:
            if self.output_file is not None:
                return self.output_file
            else:
                from bokeh.io.util import detect_current_filename

                return detect_current_filename()

    @property
    def source(self):
        """

        :return:
        """
        from bokeh.models import ColumnDataSource

        return ColumnDataSource

    def draw(self):
        """

        :return:
        """
        if not self.server:
            plt = self.plt
            if self.output_notebook:
                plt.output_notebook()
            else:
                if self.output_file is not None:
                    path = pathlib.Path(self.output_file).parent.absolute()
                    if not path.is_dir():
                        path.mkdir(parents=True, exist_ok=True)
                    plt.output_file(self.output_file)
            plt.show(self.axes, browser=self.browser)

    def generate(self):
        """

        :return:
        """
        self._setup_subplots()
        self._make_plot()
        self._make_fill_between()
        self._make_legend()
        self._decorate_subplots()


class MultiPlot(BasePlot):
    """"""
    def __init__(self, data, kind, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

        self.kind = kind
        if len(self.kind) != self.n_plots:
            msg = f"The length of 'kind' must equal the number of columns if using 'kind' of type " \
                  f"'list'.\nlength of kind = {len(self.kind)}\nnumber of columns = {self.n_plots}"
            raise ValueError(msg)

        self._organize_kwargs()

    def _organize_kwargs(self):
        """

        :return:
        """
        supplied_kinds = set()
        for kind in self.kind:
            if is_list_like(kind):
                supplied_kinds.update(kind)
            else:
                supplied_kinds.add(kind)

        self._line_kwargs = {}
        self._step_kwargs = {}
        self._scatter_kwargs = {}

        for kw, val in self.kwargs.items():
            if kw in _LINE_KWARGS:
                if 'line' in supplied_kinds:
                    self._line_kwargs[kw] = val
                if 'dashed' in supplied_kinds:
                    self._line_kwargs[kw] = val
                if 'dotted' in supplied_kinds:
                    self._line_kwargs[kw] = val
                if 'dotdash' in supplied_kinds:
                    self._line_kwargs[kw] = val
                if 'dashdot' in supplied_kinds:
                    self._line_kwargs[kw] = val
                if 'step' in supplied_kinds:
                    self._step_kwargs[kw] = val
            if kw in _STEP_KWARGS:
                if 'step' in supplied_kinds:
                    if kw == 'step_mode':
                        self._step_kwargs['mode'] = val
                    else:
                        self._step_kwargs[kw] = val
            if kw in _SCATTER_KWARGS:
                if 'scatter' in supplied_kinds:
                    if kw == 'marker_size':
                        self._scatter_kwargs['size'] = val
                    else:
                        self._scatter_kwargs[kw] = val
            if kw == 'x_range':
                self._line_kwargs[kw] = val
                self._step_kwargs[kw] = val
                self._scatter_kwargs[kw] = val

    def _get_kwargs(self, kind):
        """

        :return:
        """
        if kind in ['line', 'dashed', 'dotted', 'dotdash', 'dashdot']:
            return self._line_kwargs.copy()
        elif kind == 'step':
            return self._step_kwargs.copy()
        elif kind == 'scatter':
            return self._scatter_kwargs.copy()
        else:
            return self.kwargs.copy()

    def _iter_data(self):
        """

        :return:
        """
        for sub, plots in self.data.items():
            if len(self.sources) < self.n_plots:
                sources = []
                if len(plots['x']) == len(plots['y']) == len(self.kind[sub]):
                    for key, val in plots['x'].items():
                        source_data = {}
                        source_data['x'] = val
                        source_data['y0'] = plots['y'][key]
                        source = self.source(source_data)
                        sources.append(source)
                elif len(plots['x']) == 1 and len(plots['y']) == len(self.kind[sub]):
                    raise NotImplementedError
                elif len(plots['x']) == len(plots['y']) and len(self.kind[sub]) == 1:
                    raise NotImplementedError
                elif len(plots['x']) == len(self.kind[sub]) == 1:
                    raise NotImplementedError
                else:
                    raise ValueError("Dimension mismatch between the number of plot elements for both axes and the "
                                     "supplied plot types.")
                self.sources.append(sources)
            elif len(self.sources) == self.n_plots:
                sources = self.sources[sub]
            else:
                raise RuntimeError("length of ColumnDataSource list > number of plots")

            colors = self._get_colors(num_colors=len(plots['y']))
            keys = list(plots['y'])  # TODO: Check if len(plots['y']) and len(self.kind[sub]) are the same. The code
            # TODO: needs to be adjusted if this is not the case.
            for k, kind in enumerate(self.kind[sub]):
                yield sub, k, [keys[k]], sources[k], colors[k], kind

    def _make_plot(self):
        """

        :return:
        """
        it = self._iter_data()

        for (subplot, k, labels, source, colors, kind) in it:
            ax = self._get_glyph(subplot)
            kwargs = self._get_kwargs(kind)

            if kind == 'scatter':
                if is_list_like(kwargs['marker']):
                    kwargs['marker'] = kwargs['marker'][subplot][k]
                if is_list_like(kwargs['size']):
                    kwargs['size'] = kwargs['size'][subplot][k]
            # NOTE: Going with list(color) for now since processing string representation is not yet implemented
            kwargs['color'] = colors
            kwargs['legend'] = True
            kwargs['ax'] = ax
            kwargs['sources'] = [source]

            data = {
                0: {
                    'x': [],
                    'y': {label: [] for label in labels}
                }
            }

            plot_obj = MultiPlot.return_plot_class(kind)(data, **kwargs)
            plot_obj.generate()

    @staticmethod
    def return_plot_class(arg):
        """

        :param arg:
        :return:
        """
        return {
            'line': LinePlot,
            'dashed': DashedPlot,
            'dotted': DottedPlot,
            'dotdash': DotDashPlot,
            'dashdot': DashDotPlot,
            'step': StepPlot,
            'scatter': ScatterPlot
        }[arg]


class LinePlot(BasePlot):
    """"""
    def __init__(self, data, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

    def _make_plot(self):
        """

        :return:
        """
        it = self._iter_data()

        for (subplot, k, label, source, color) in it:
            ax = self._get_glyph(subplot)
            kwargs = self.kwargs.copy()

            x_range = kwargs.pop('x_range', None)

            kwargs['line_color'] = color
            kwargs['legend_label'] = label
            kwargs['source'] = source

            ax.line(x='x', y='y' + str(k), **kwargs)

            if x_range is None:
                left, right = _get_x_lim(ax.renderers)
            else:
                left, right = x_range
            ax.x_range.start = left
            ax.x_range.end = right


class DashedPlot(LinePlot):
    """"""
    def __init__(self, data, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

        self.kwargs['line_dash'] = 'dashed'


class DottedPlot(LinePlot):
    """"""
    def __init__(self, data, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

        self.kwargs['line_dash'] = 'dotted'


class DotDashPlot(LinePlot):
    """"""
    def __init__(self, data, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

        self.kwargs['line_dash'] = 'dotdash'


class DashDotPlot(LinePlot):
    """"""
    def __init__(self, data, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

        self.kwargs['line_dash'] = 'dashdot'


class StepPlot(BasePlot):
    """"""
    def __init__(self, data, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

    def _make_plot(self):
        """

        :return:
        """
        it = self._iter_data()

        for (subplot, k, label, source, color) in it:
            ax = self._get_glyph(subplot)
            kwargs = self.kwargs.copy()

            x_range = kwargs.pop('x_range', None)

            kwargs['line_color'] = color
            kwargs['legend_label'] = label
            kwargs['source'] = source

            ax.step(x='x', y='y' + str(k), **kwargs)

            if x_range is None:
                left, right = _get_x_lim(ax.renderers)
            else:
                left, right = x_range
            ax.x_range.start = left
            ax.x_range.end = right


class ScatterPlot(BasePlot):
    """"""
    def __init__(self, data, marker='x', **kwargs):
        """Constructor class"""
        super().__init__(data, **kwargs)

        self.marker = marker
        if is_list_like(self.marker) and len(self.marker) != self.n_plots:
            msg = f"The length of 'marker' must equal the number of columns if using 'marker' of type " \
                  f"'list'.\nlength of marker = {len(self.marker)}\nnumber of columns = {self.n_plots}"
            raise ValueError(msg)

    def _make_plot(self):
        """

        :return:
        """
        it = self._iter_data()

        for (subplot, k, label, source, color) in it:
            ax = self._get_glyph(subplot)
            kwargs = self.kwargs.copy()

            x_range = kwargs.pop('x_range', None)

            kwargs['color'] = color
            kwargs['legend_label'] = label
            kwargs['source'] = source

            # NOTE: k should actually always be zero at the moment, since we only put one plottable in
            #  ColumnDataSource to better handle different dimension
            if is_list_like(self.marker):
                marker = self.marker[subplot]
                if is_list_like(marker):
                    marker = marker[k]
            else:
                marker = self.marker
            marker_size = kwargs.get('size')
            if marker_size is None:
                marker_size = kwargs.pop('marker_size', None)
            if marker_size is not None:
                if is_list_like(marker_size):
                    marker_size = marker_size[subplot]
                    if is_list_like(marker_size):
                        kwargs['size'] = marker_size[k]
                    else:
                        kwargs['size'] = marker_size

            self._plot(ax, marker)(x='x', y='y' + str(k), **kwargs)

            if x_range is None:
                left, right = _get_x_lim(ax.renderers)
            else:
                left, right = x_range
            ax.x_range.start = left
            ax.x_range.end = right

    @staticmethod
    def _plot(ax, marker):
        """

        :param ax:
        :param marker:
        :return:
        """
        return {
            '*': ax.asterisk,
            'o': ax.circle,
            'o+': ax.circle_cross,
            'o.': ax.circle_dot,
            'ox': ax.circle_x,
            'oy': ax.circle_y,
            '+': ax.cross,
            '-': ax.dash,
            'd': ax.diamond,
            'd+': ax.diamond_cross,
            'd.': ax.diamond_dot,
            '.': ax.dot,
            'h': ax.hex,
            'h.': ax.hex_dot,
            'it': ax.inverted_triangle,
            'pl': ax.plus,
            'sq': ax.square,
            'sq+': ax.square_cross,
            'sq.': ax.square_dot,
            'sqp': ax.square_pin,
            'sqx': ax.square_x,
            's': ax.star,
            's.': ax.star_dot,
            'tr': ax.triangle,
            'tr.': ax.triangle_dot,
            'trp': ax.triangle_pin,
            'x': ax.x,
            'y': ax.y
        }[marker]


def _convert_color(c):
    """

    :param c:
    :return:
    """
    if c not in 'bgrcmykw':
        return c
    return {
        'b': 'blue',
        'g': 'green',
        'r': 'red',
        'c': 'cyan',
        'm': 'magenta',
        'y': 'yellow',
        'k': 'black',
        'w': 'white'
    }[c]


def _get_layout(n_plots, layout=None, layout_type='grid'):
    """

    :param n_plots:
    :param layout:
    :param layout_type:
    :return:
    """
    if layout is not None:
        if not isinstance(layout, (list, tuple)) or len(layout) != 2:
            raise ValueError("Layout must be a tuple of (rows, columns)")

        n_rows, n_cols = layout

        def ceil_(x):
            """

            :param x:
            :return:
            """
            return int(ceil(x))

        if n_rows == -1 and n_cols > 0:
            layout = n_rows, n_cols = (ceil_(float(n_plots) / n_cols), n_cols)
        elif n_cols == -1 and n_rows > 0:
            layout = n_rows, n_cols = (n_rows, ceil_(float(n_plots) / n_rows))
        elif n_cols <= 0 and n_rows <= 0:
            msg = "At least one dimension of layout must be positive"
            raise ValueError(msg)

        if n_rows * n_cols < n_plots:
            raise ValueError(f"Layout of {n_rows}x{n_cols} must be larger than required size {n_plots}")

        return layout

    if layout_type == 'single':
        return 1, 1
    if layout_type == 'horizontal':
        return 1, n_plots
    if layout_type == 'vertical':
        return n_plots, 1

    layouts = {1: (1, 1), 2: (1, 2), 3: (2, 2), 4: (2, 2)}
    try:
        return layouts[n_plots]
    except KeyError:
        k = 1
        while k ** 2 < n_plots:
            k += 1

        if (k - 1) * k >= n_plots:
            return (k - 1), k
        else:
            return k, k


def _get_standard_colors(num_colors=None, colormap=None, color_type='default', color=None):
    """

    :param num_colors:
    :param colormap:
    :param color_type:
    :param color:
    :return:
    """
    if color is None and colormap is not None:
        if isinstance(colormap, str):
            cmap = colormap
            if 'Category' in colormap and not colormap.isdigit():
                # TODO: Maybe we could combine all Category20's if we have '<= 60' colors?
                if num_colors <= 10:
                    colormap += '10'
                else:
                    colormap += '20'
            if colormap in cm.all_palettes:
                colormap = cm.all_palettes[colormap]
            else:
                raise ValueError(f"Colormap {cmap} is not recognized")
        last = list(colormap)[-1]
        colors = [colormap[last][k] for k in range(num_colors)]
    elif color is not None:
        if colormap is not None:
            warnings.warn("'color' and 'colormap' cannot be used simultaneously. Using 'color'...")
        colors = (list(color) if is_list_like(color) and not isinstance(color, dict) else color)
    else:
        if color_type == 'default':
            if num_colors <= 10:
                colors = cm.Category10_10
            else:
                colors = cm.Category20_20

            colors = colors[:num_colors]
        elif color_type == 'random':
            def random_color(column):
                """

                :param column:
                :return:
                """
                rs = random_state(column)
                return rs.rand(3).to_list()

            colors = [random_color(k) for k in range(num_colors)]
        else:
            raise ValueError("'color_type' must be either 'default' or 'random'")

    if isinstance(colors, str):
        from bokeh.colors import named

        def _maybe_valid_colors(colors):
            """

            :param colors:
            :return:
            """
            try:
                [getattr(named, _convert_color(c)) for c in colors]
                return True
            except AttributeError:
                return False

        maybe_single_color = _maybe_valid_colors([colors])
        maybe_color_cycle = _maybe_valid_colors(list(colors))
        if maybe_single_color and maybe_color_cycle and len(colors) > 1:
            # NOTE: Some HEX stuff
            raise NotImplementedError
        elif maybe_color_cycle:
            if len(colors) > 1:
                colors = [_convert_color(c) for c in colors]
            else:
                colors = [_convert_color(colors)]
        else:
            colors = [colors]

    if len(colors) < num_colors:
        try:
            multiple = num_colors // len(colors) - 1
        except ZeroDivisionError:
            raise ValueError("Invalid color argument: ''")
        mod = num_colors % len(colors)

        colors += multiple * colors
        colors += colors[:mod]

    return colors


def _get_x_lim(lines):
    """

    :param lines:
    :return:
    """
    left, right = np.inf, -np.inf
    for line in lines:
        x = line.data_source.data['x']
        left = min(np.nanmin(x), left)
        right = max(np.nanmax(x), right)
    return left, right


def _subplots(n_axes=None, subplot_kwargs=None, layout=None, layout_type='grid', **kwargs):
    """

    :param n_axes:
    :param subplot_kwargs:
    :param layout:
    :param layout_type:
    :param kwargs:
    :return:
    """
    from bokeh.plotting import figure
    from bokeh.layouts import column, row, gridplot

    if subplot_kwargs is None:
        subplot_kwargs = {}
        x_data = None
    else:
        interactive = subplot_kwargs.pop('interactive', False)
        if interactive:
            x_data = subplot_kwargs.pop('x_data', None)
            if x_data is None:
                raise KeyError("Argument 'x_data' is missing. If interactive flag is set to True, 'x_data' needs to be"
                               " supplied as well")
        else:
            x_data = None

    n_rows, n_cols = _get_layout(n_axes, layout=layout, layout_type=layout_type)
    n_plots = n_rows * n_cols

    fig = figure(**subplot_kwargs)
    if x_data is not None:
        fig = column(fig, _slider(x_data[0]))

    if n_rows == 1 and n_cols > 1:
        figs = [fig]
        for k in range(1, n_plots):
            if k < n_axes:
                fig = figure(**subplot_kwargs)
                if x_data is not None:  # Means, that interactive flag was set to True
                    fig = column(fig, _slider(x_data[k]))
                figs.append(fig)
        return row(*figs, **kwargs)
    elif n_cols == 1 and n_rows > 1:
        figs = [fig]
        for k in range(1, n_plots):
            if k < n_axes:
                fig = figure(**subplot_kwargs)
                if x_data is not None:  # Means, that interactive flag was set to True
                    fig = column(fig, _slider(x_data[k]))
                figs.append(fig)
        return column(*figs, **kwargs)
    elif n_rows > 1 and n_cols > 1:
        grid = []
        for i in range(n_rows):
            row = []
            for j in range(n_cols):
                if i + j == 0:
                    row.append(fig)
                    continue
                if i * n_cols + j < n_axes:
                    fig = figure(**subplot_kwargs)
                    if x_data is not None:  # Means, that interactive flag was set to True
                        fig = column(fig, _slider(x_data[i * n_cols + j]))
                    row.append(fig)
                else:
                    row.append(None)
            grid.append(row)
        return gridplot(grid, **kwargs)
    else:
        return fig


def _slider(data: dict[str, np.ndarray], epsilon: float = 1e-8):
    """

    :param data:
    :param epsilon:
    :return:
    """
    from bokeh.models import Slider

    expected = next(iter(data.values()))
    all_equal = all(
        (value == expected).all() if value.size == expected.size else value == expected for value in data.values())
    if all_equal:
        x_data = expected
    else:
        index = np.argmax(value.size for value in data.values())
        x_data = list(data.values())[index]
    n_data = len(x_data)
    diff = np.diff(x_data)
    unique_diff = np.all(np.abs(diff - diff[0]) < epsilon)

    if unique_diff:
        slider = Slider(start=x_data[0], end=x_data[-1], value=x_data[-1], step=diff[0])
    else:
        slider = Slider(start=0, end=n_data - 1, value=n_data - 1, step=1)

    return slider


def _update_slider(slider, sources):
    """

    :param slider:
    :param sources:
    :return:
    """
    from bokeh.models import ColumnDataSource, CustomJS

    full_sources = [ColumnDataSource(source.data) for source in sources]

    n_y = len(sources)
    n_data = [source.data['x'].size for source in sources]

    callback = CustomJS(args=dict(source=sources, full_source=full_sources, n_y=n_y, n_data=n_data), code="""
        var val = cb_obj.value;
        for (var i=0; i<n_y; i++) {
            var data = source[i].data;
            var full_data = full_source[i].data;
            var x = data['x'];
            var y = data['y0'];
            var y_full = full_data['y0'];
            for (var j=0; j<n_data[i]; j++) {
                if (x[j]>val) {
                    y[j] = Number.NaN;
                } else {
                    y[j] = y_full[j];
                }
            }
            source[i].change.emit();
        }
    """)

    slider.js_on_change('value', callback)


__all__ = [
    'MultiPlot',
    'LinePlot',
    'StepPlot',
    'DashedPlot',
    'DottedPlot',
    'DotDashPlot',
    'DashDotPlot',
    'ScatterPlot'
]






#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from typing import Optional
import uuid


class Object:
    """
    Defines common behavior of various objects in HILO-MPC.

    :param id: Identifier of the object
    :type id: str, optional
    :param name: Name of the object
    :type name: str, optional
    """
    def __init__(self, id: Optional[str] = None, name: Optional[str] = None) -> None:
        """Constructor method"""
        self._id = id
        self.name = name

    def __repr__(self) -> str:
        """Representation method"""
        args = ""
        if self._id is not None:
            args += f"id='{self._id}'"
        if self.name is not None:
            args += f", name='{self.name}'"
        return f"{self.__class__.__name__}({args})"

    def __str__(self) -> str:
        """String representation method"""
        return f"{self.__class__.__name__}: {self.name} ({self._id})"

    def _create_id(self) -> None:
        """
        Generates a unique identifier for the object.

        :return:
        """
        self._id = str(uuid.uuid4())

    @property
    def id(self) -> str:
        """
        Identifier of the object

        :return: Identifier of the object
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        if value != self._id:
            if not isinstance(value, str):
                raise TypeError(f"{self.__class__.__name__} ID must be a string")
            else:
                self._id = value



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from distutils.version import StrictVersion
import os


_PYTORCH_VERSION = '1.2.0'
_TENSORFLOW_VERSION = ('2.3.0', '2.8.0')
_SCIKIT_LEARN_VERSION = '0.19.2'
_TENSORBOARD_VERSION = '2.3.0'
_BOKEH_VERSION = '2.3.0'
_MATPLOTLIB_VERSION = '3.0.0'
_PANDAS_VERSION = '2.0.0'


class Manager:
    """
    Base class for managers

    :param backend:
    :type backend:
    """
    # TODO: Typing hints
    def __init__(self, backend):
        """Constructor method"""
        self._backend = backend

    @property
    def backend(self):
        """

        :return:
        """
        return self._backend

    @backend.setter
    def backend(self, arg):
        self._backend = arg


class LearningManager(Manager):
    """"""
    def setup(self, kind, *args, **kwargs):
        """

        :param kind:
        :param args:
        :param kwargs:
        :return:
        """
        if isinstance(self._backend, str):
            backend = _get_learning_backend(self._backend)
        else:
            backend = self._backend
        return backend.setup(kind, *args, **kwargs)


class LearningVisualizationManager(Manager):
    """"""
    def setup(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        if isinstance(self._backend, str):
            backend = _get_learning_visualization_backend(self._backend)
        else:
            backend = self._backend
        return backend.setup(*args, **kwargs)


class PlotManager(Manager):
    """"""
    def plot(self, data, kind='line', **kwargs):
        """

        :param data:
        :param kind:
        :param kwargs:
        :return:
        """
        if isinstance(self._backend, str):
            plot_backend = _get_plot_backend(self._backend)
        else:
            if not hasattr(self._backend, 'plot'):
                raise AttributeError(f"Supplied backend '{self._backend}' has no attribute 'plot'")
            plot_backend = self._backend
        return plot_backend.plot(data, kind, **kwargs)


def _get_learning_backend(backend):
    """

    :param backend:
    :return:
    """
    if backend == 'pytorch':
        try:
            import torch
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Backend 'PyTorch' is not installed")
        if StrictVersion(torch.__version__.split('+')[0]) < StrictVersion(_PYTORCH_VERSION):
            raise RuntimeError(f"Backend 'PyTorch' is installed with version '{torch.__version__}', but version "
                               f"'{_PYTORCH_VERSION}' or higher is required")
        import hilo_mpc.plugins.pytorch as module
    elif backend == 'tensorflow':
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
        try:
            import tensorflow
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Backend 'TensorFlow' is not installed")
        if StrictVersion(tensorflow.__version__) < StrictVersion(_TENSORFLOW_VERSION[0]) or \
                StrictVersion(tensorflow.__version__) >= StrictVersion(_TENSORFLOW_VERSION[1]):
            raise RuntimeError(f"Backend 'TensorFlow' is installed with version '{tensorflow.__version__}', but version"
                               f" needs to be higher or equal to '{_TENSORFLOW_VERSION[0]}' and lower than "
                               f"'{_TENSORFLOW_VERSION[1]}'")
        import hilo_mpc.plugins.tensorflow as module
    elif backend in ['sklearn', 'scikit-learn']:
        try:
            import sklearn
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Backend 'scikit-learn' is not installed")
        if StrictVersion(sklearn.__version__) < StrictVersion(_SCIKIT_LEARN_VERSION):
            raise RuntimeError(f"Backend 'scikit-learn' is installed with version '{sklearn.__version__}', but version"
                               f" '{_SCIKIT_LEARN_VERSION}' or higher is required")
        import hilo_mpc.plugins.sklearn as module
    else:
        raise ValueError(f"Backend '{backend}' not recognized")
    return module


def _get_learning_visualization_backend(backend):
    """

    :param backend:
    :return:
    """
    if backend == 'tensorboard':
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
        try:
            import tensorboard
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Backend 'Tensorboard' is not installed")
        import hilo_mpc.plugins.tensorboard as module
    else:
        raise ValueError(f"Backend '{backend}' not recognized")
    return module


def _get_plot_backend(backend):
    """

    :param backend:
    :return:
    """
    if backend == 'matplotlib':
        try:
            import matplotlib
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Backend 'Matplotlib' is not installed")
        if StrictVersion(matplotlib._get_version()) < StrictVersion(_MATPLOTLIB_VERSION):
            raise RuntimeError(f"Backend 'Matplotlib' is installed with version '{matplotlib._get_version()}', but "
                               f"version '{_MATPLOTLIB_VERSION}' or higher is required")
        import hilo_mpc.plugins.matplotlib as module
    elif backend == 'bokeh':
        try:
            import bokeh
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Backend 'Bokeh' is not installed")
        if StrictVersion(bokeh.__version__) < StrictVersion(_BOKEH_VERSION):
            raise RuntimeError(f"Backend 'Bokeh' is installed with version '{bokeh.__version__}', but version "
                               f"'{_BOKEH_VERSION}' or higher is required")
        import hilo_mpc.plugins.bokeh as module
    else:
        raise ValueError(f"Backend '{backend}' not recognized")

    return module


def check_version(library: str) -> None:
    """

    :param library:
    :return:
    """
    if library == 'pandas':
        try:
            import pandas
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Plugin 'pandas' is not installed")
        if StrictVersion(pandas.__version__) < StrictVersion(_PANDAS_VERSION):
            raise RuntimeError(f"Plugin 'pandas' is installed with version '{pandas.__version__}', but version "
                               f"{_PANDAS_VERSION}' or higher is required")
